package com.loafle.overflow.probe.container.rpc.invoker.spring;

import com.fasterxml.jackson.databind.JavaType;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.loafle.overflow.probe.container.rpc.invoker.Invoker;
import com.loafle.overflow.probe.container.util.Primitives;
import org.springframework.aop.support.AopUtils;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationContext;
import org.springframework.stereotype.Component;

import java.io.IOException;
import java.lang.reflect.Method;
import java.lang.reflect.Type;
import java.security.InvalidParameterException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

@Component
public class SpringInvoker implements Invoker {
    @Autowired
    private ApplicationContext applicationContext;

    @Autowired
    private ObjectMapper objectMapper;

    private Map<String, Cache> serviceCacheMap;

    public SpringInvoker() {
        serviceCacheMap = new HashMap<>();
    }

    private static class Cache {
        private ServiceCache serviceCache;
        private Map<String, MethodCache> methodCacheMap;

        private Cache(Object bean) {
            serviceCache = new ServiceCache();
            serviceCache.bean = bean;

            methodCacheMap = new HashMap<>(10);
        }

        private Object getBean() {
            return serviceCache.bean;
        }

        private MethodCache getMethodCache(String methodName) throws NoSuchMethodException {
            MethodCache methodCache = methodCacheMap.get(methodName);
            if (null != methodCache) {
                return methodCache;
            }

            Method method = getMethod(methodName);
            if (null == method) {
                throw new NoSuchMethodException();
            }
            methodCache = new MethodCache();
            methodCacheMap.put(methodName, methodCache);

            methodCache.method = method;
            methodCache.returnClazz = method.getReturnType();

            int paramCount = method.getParameterCount();
            if (0 == paramCount) {
                return methodCache;
            }

            methodCache.parameterTypes = method.getGenericParameterTypes();

            return methodCache;
        }

        private Method getMethod(String methodName) {
            Class<?> clazz = AopUtils.getTargetClass(serviceCache.bean);
            Method[] methods = clazz.getMethods();

            Method targetMethod = null;
            for(Method m : methods){
                if (methodName.equals(m.getName())) {
                    targetMethod = m;
                    break;
                }
            }
            return targetMethod;
        }

        private static class ServiceCache {
            private Object bean;
            private Map<String, MethodCache> methodCacheMap;
        }
        private static class MethodCache {
            private Method method;
            private Class returnClazz;
            private Type[] parameterTypes;
        }
    }

    private Cache getServiceCache(String serviceName) throws Exception {
        Cache serviceCache = serviceCacheMap.get(serviceName);
        if (null != serviceCache) {
            return serviceCache;
        }
        try {
            Object bean = applicationContext.getBean(serviceName);
            serviceCache = new Cache(bean);
            serviceCacheMap.put(serviceName, serviceCache);
        } catch (BeansException e) {
            throw new Exception();
        }

        return serviceCache;
    }

    private Object getValue(Type parameterType, String json) throws InvalidParameterException {
        JavaType targetType = objectMapper.getTypeFactory().constructType(parameterType);
        if (!Primitives.isPrimitive(parameterType) && !parameterType.getTypeName().equals(String.class.getName())) {
            try {
                return objectMapper.readValue(json, targetType);
            } catch (IOException e) {
                throw new InvalidParameterException();
            }
        }
        return objectMapper.convertValue(json, targetType);
    }


    private Object[] getParameters(Type[] parameterTypes, List<String> params) throws InvalidParameterException {
        if (null == parameterTypes || null == params) {
            return null;
        }

        if (parameterTypes.length != params.size()) {
            throw new IllegalArgumentException();
        }

        Object[] result = new Object[parameterTypes.length];

        for (int i = 0; i < parameterTypes.length; i++) {
            result[i] = getValue(parameterTypes[i], params.get(i));
        }

        return result;
    }


    public Object invoke(String method, List<String> params) throws Exception {
        String[] ms = method.split("\\.");
        String serviceName = ms[0];
        String methodName = ms[1];

        Cache serviceCache = getServiceCache(serviceName);
        Cache.MethodCache methodCache = serviceCache.getMethodCache(methodName);

        if (null == methodCache.parameterTypes) {
            if (null != params && 0 < params.size()) {
                throw new InvalidParameterException();
            }
        } else {
            if (methodCache.parameterTypes.length != params.size()) {
                throw new InvalidParameterException();
            }
        }

        Object[] parameters = null;
        if (null != methodCache.parameterTypes && 0 < methodCache.parameterTypes.length) {
            parameters = getParameters(methodCache.parameterTypes, params);
        }

        Object result = null;
        try {
            result = methodCache.method.invoke(serviceCache.getBean(), parameters);
        } catch (Exception e) {
            throw e;
        }

        return result;
    }
}
